from typing import Dict, Any, Optional, List, Callable, Coroutine

import aiohttp
import orjson
import logging

from linkauto.api.enums import TimestampType
from linkauto.api.exceptions import BadRequestException, BadPasswordException
from linkauto.api.mixins.stub import StubMixin

logger = logging.getLogger(__name__)


class HttpMixin(StubMixin):
  _session: aiohttp.ClientSession
  _response_interceptors: List[Callable[[aiohttp.ClientResponse], Coroutine]] = []

  def start_http_mixin(self):
    self._session = aiohttp.ClientSession(cookies={'JSESSIONID': self.generate_csrf_token()})
    self.add_response_interceptor(self._error_handler)

  async def stop_http_mixin(self):
    await self._session.close()

  def add_response_interceptor(self, c: Callable[[aiohttp.ClientResponse], Coroutine]):
    self._response_interceptors.append(c)

  def get_csrf_token(self) -> str:
    return self._session.cookie_jar._cookies['']['JSESSIONID'].coded_value

  def get_lidc_token(self) -> str:
    return self._session.cookie_jar._cookies['linkedin.com']['lidc'].coded_value

  def get_bcookie(self) -> str:
    return self._session.cookie_jar._cookies['linkedin.com']['bcookie'].coded_value

  def get_bscookie(self) -> str:
    return self._session.cookie_jar._cookies['www.linkedin.com']['bscookie'].coded_value

  async def _request(self,
                     url: str,
                     method: str,
                     data: Optional[Any] = None,
                     query_params: Optional[Dict[Any, Any]] = None,
                     headers: Optional[Dict[str, str]] = None,
                     default_headers: bool = True,
                     default_headers_overwrite: bool = False,
                     accept_json: bool = False,
                     page_instance: Optional[str] = None,
                     send_json: bool = False
                     ) -> aiohttp.ClientResponse:
    url = self._prepare_url(url)
    h: Dict[str, str] = self._build_headers(headers, default_headers, default_headers_overwrite)
    query_params = self._build_query_params(query_params or {})
    if page_instance is not None:
      h['x-li-page-instance'] = self.get_page_instance(page_instance)
    if accept_json:
      h['accept'] = 'application/vnd.linkedin.mobile.deduped+json'
    if method.upper() == 'GET':
      if data is not None:
        logger.warning("You're sending a get request with data. The data will be ignored")
      response = await self._session.get(url, headers=headers, params=query_params)
    elif method.upper() == 'POST':
      if send_json:
        data = orjson.dumps(data)
        h['content-type'] = 'application/json'
      response = await self._session.post(url, data=data, headers=headers, params=query_params)
    else:
      raise ValueError('Invalid method: %s', method)
    _ = (await c(response) for c in self._response_interceptors)
    return response

  async def paginate(self, response: Optional[aiohttp.ClientResponse] = None, step: int = 10) -> aiohttp.ClientResponse:
    data = orjson.loads(await response.read())
    if isinstance(data, list):
      raise ValueError("Data is a list. Paginating is not possible.")

    paging = data['paging']
    url = response.real_url.path
    query_params = {
      'count': step,
      'start': paging['start'] + step
    }

    method = response.method
    return await self._request(
      url,
      method,
      query_params=query_params,
      headers={k.decode(): v.decode() for k, v in response.raw_headers},
      default_headers_overwrite=True
    )

  async def get(self,
                url: str,
                query_params: Optional[Dict[Any, Any]] = None,
                headers: Optional[Dict[str, str]] = None,
                default_headers: bool = True,
                accept_json: bool = False,
                page_instance: Optional[str] = None,
                ) -> aiohttp.ClientResponse:
    return await self._request(url, 'GET', query_params=query_params, headers=headers, default_headers=default_headers,
                               accept_json=accept_json, page_instance=page_instance)

  async def post(self, url: str,
                 data: Optional[Any] = None,
                 query_params: Optional[Dict[Any, Any]] = None,
                 headers: Optional[Dict[str, str]] = None,
                 default_headers: bool = True,
                 accept_json: bool = False,
                 page_instance: Optional[str] = None,
                 send_json: bool = False
                 ) -> aiohttp.ClientResponse:
    return await self._request(url, 'POST', query_params=query_params, headers=headers, default_headers=default_headers,
                               accept_json=accept_json, page_instance=page_instance, data=data, send_json=send_json)

  def _build_headers(self, headers: Optional[Dict[str, str]], default_headers: bool, default_headers_overwrite: bool):
    if not default_headers_overwrite:
      return {**(self._get_default_headers() if default_headers else {}), **(headers or {})}
    return {**(headers or {}), **(self._get_default_headers() if default_headers else {})}

  def _build_query_params(self, query_params: Dict[Any, Any]) -> Dict[Any, Any]:
    query_params['nc'] = self.get_timestamp(TimestampType.milliseconds)
    return query_params

  def _get_default_headers(self) -> Dict[str, str]:
    return {
      'user-agent': self.get_useragent(),
      'x-udid': str(self.config.linkedin_stored_data.li_udid),
      'csrf-token': self.get_csrf_token(),
      'x-restli-protocol-version': '2.0.0',
      'x-li-track': self.get_track_header(),
      'x-li-lang': self.config.linkedin_stored_data.li_lang
    }

  def _prepare_url(self, url: str) -> str:
    return "https://www.linkedin.com/" + url.lstrip('/')

  async def _error_handler(self, response: aiohttp.ClientResponse):
    if response.ok:
      return

    try:
      as_json: dict = orjson.loads(await response.read())
    except orjson.JSONDecodeError:
      raise BadRequestException("Unknown error.")

    if isinstance(as_json, list):
      return

    if login_result := as_json.get('login_result'):
      if login_result == 'BAD_PASSWORD':
        raise BadPasswordException()




from typing import Optional

import aiohttp

from linkauto.api.enums import ConnectionInviteType
from linkauto.api.mixins.stub import StubMixin


class RelationshipMixin(StubMixin):
  async def relationship_invitation_create(self,
                                           message: Optional[str] = None,
                                           profile_url: Optional[str] = None,
                                           id_: Optional[str] = None
                                           ) -> aiohttp.ClientResponse:
    if not any([profile_url, id_]):
      raise ValueError("Neither an id, or an url has been provided.")

    if id_ is None:
      id_ = await self.profile_url_to_id(profile_url)

    data = {
      'trackingId': self.get_tracking_id(),
      'invitee': {
        'com.linkedin.voyager.growth.invitation.InviteeProfile': {
          'profileId': id_
        }
      },
    }
    if message is not None:
      data['message'] = message

    return await self.post(
      f'/voyager/api/voyagerGrowthNormInvitations',
      data=data,
      page_instance='urn:li:page:p_flagship3_profile_view_base',
      accept_json=True,
      send_json=True
    )

  async def relationship_invitation_get(self, type_: ConnectionInviteType):
    url = "/voyager/api/relationships/invitationViews"
    if type_ == type_.SENT:
      url = "/voyager/api/relationships/sentInvitationViewsV2"
    return await self.get(
      url,
      query_params={
        'q': type_.value,
        'invitationType': 'CONNECTION',
        'start': 0,
        'count': 10
      },
      accept_json=True,
      page_instance='urn:li:page:p_flagship3_people_sent_invitations'
    )



from typing import Optional, Dict, Any, Callable, Coroutine

import aiohttp

from linkauto.api.dataclasses_ import Config
from linkauto.api.enums import TimestampType


class _get:
  async def __call__(self,
                url: str,
                query_params: Optional[Dict[Any, Any]] = None,
                headers: Optional[Dict[str, str]] = None,
                default_headers: bool = True,
                accept_json: bool = False,
                page_instance: Optional[str] = None
                ) -> aiohttp.ClientResponse: ...


class _post:
  async def __call__(self, url: str,
                 data: Optional[Any] = None,
                 query_params: Optional[Dict[Any, Any]] = None,
                 headers: Optional[Dict[str, str]] = None,
                 default_headers: bool = True,
                 accept_json: bool = False,
                 page_instance: Optional[str] = None,
                 send_json: bool = False
                 ) -> aiohttp.ClientResponse: ...


class _get_useragent:
  def __call__(self) -> str: ...


class _get_page_instance:
  def __call__(self, page: str) -> str: ...


class _get_timestamp:
  @staticmethod
  def __call__(d: TimestampType) -> int: ...


class _generate_csrf_token:
  def __call__(self) -> str: ...


class _get_csrf_token:
  def __call__(self) -> str: ...


class _add_response_interceptor:
  def __call__(self, c: Callable[[aiohttp.ClientResponse], Coroutine]): ...


class _get_track_header:
  def __call__(self) -> str: ...


class _get_lidc_token:
  def __call__(self) -> str: ...


class _get_bcookie:
  def __call__(self) -> str: ...


class _get_bscookie:
  def __call__(self) -> str: ...


class _profile_url_to_id:
  async def __call__(self, url: str) -> str: ...


class _get_tracking_id:
  def __call__(self) -> str: ...


class StubMixin:
  username: str
  password: str
  config: Config
  get: _get
  post: _post
  get_useragent: _get_useragent
  get_page_instance: _get_page_instance
  get_timestamp: _get_timestamp
  generate_csrf_token: _generate_csrf_token
  get_csrf_token: _get_csrf_token
  add_response_interceptor: _add_response_interceptor
  get_track_header: _get_track_header
  get_lidc_token: _get_lidc_token
  get_bcookie: _get_bcookie
  get_bscookie: _get_bscookie
  profile_url_to_id: _profile_url_to_id
  get_tracking_id: _get_tracking_id



import base64
import random
import uuid

from linkauto.api.enums import TimestampType

import time

from linkauto.api.mixins.stub import StubMixin


class HelperMixin(StubMixin):
  @staticmethod
  def get_timestamp(d: TimestampType) -> int:
    """Retrieve the current timestamp.
    Args:
        d: controls the precision of the returned timestamp.
    Returns: the current timestamp.
    """
    return int(str(time.time_ns())[:d.value])

  def get_useragent(self) -> str:
    return f"{self.config.linkedin_app.appId}/{self.config.linkedin_app.clientMinorVersion} " \
           f"(Linux; U; {self.config.device.androidVersion}; {self.config.device.region}; " \
           f"{self.config.device.model}; Build/NMF26Q; Cronet/{self.config.device.cronetVersion})"

  def get_tracking_id(self) -> str:
    return base64.b64encode(uuid.uuid4().bytes).decode()

  def get_page_instance(self, page: str) -> str:
    return f"{page};{base64.b64encode(uuid.uuid4().bytes)}"

  def generate_csrf_token(self) -> str:
    r = list(str(self.get_timestamp(TimestampType.nanoseconds)))
    random.shuffle(r)
    return f'ajax:{"".join(r)}'



import enum


class TimestampType(enum.Enum):
  seconds = 11
  milliseconds = 13
  microseconds = 15
  nanoseconds = 19


class InviteActionType(enum.Enum):
  ACCEPT = 0
  IGNORE = 1
  REJECT = 2
  REPORT_SPAM = 3
  WITHDRAW = 4
  ACTOR_WITHDRAW = 5


class ConnectionInviteType(enum.Enum):
  RECEIVED = 'receivedInvitation'
  SENT = 'invitationType'



import aiohttp

from linkauto.api.enums import TimestampType
from linkauto.api.mixins.stub import StubMixin


class AuthMixin(StubMixin):
  async def login(self) -> aiohttp.ClientResponse:
    return await self.post(
      'uas/authenticate',
      data={
        'session_key': self.username,
        'session_password': self.password,
        'client_enabled_features': 'ANDROID_NATIVE_CAPTCHA',
        'rememberMeOptIn': 'false',
        'JSESSIONID': self.get_csrf_token(),
        'lidc': self.get_lidc_token(),
        'bcookie': self.get_bcookie(),
        'bscookie': self.get_bscookie(),
        'lang': f'v=2&lang={self.config.linkedin_stored_data.li_lang}',
      },
      headers={
        'x-li-user-agent': f'LIAuthLibrary:0.0.3 {self.config.linkedin_app.appId}:{self.config.linkedin_app.clientVersion}'
                           f' {self.config.device.model}:{self.config.device.androidVersion}',
        'user-agent': 'ANDROID OS'
      }
    )

  async def logout(self) -> aiohttp.ClientResponse:
    await self.post(
      'voyager/api/pushRegistration',
      query_params={
        'action': 'deregister',
      }, headers={
        'x-li-user-agent': f'LIAuthLibrary:0.0.3 {self.config.linkedin_app.appId}:{self.config.linkedin_app.clientVersion}'
                           f' {self.config.device.model}:{self.config.device.androidVersion}'
      }
    )
    return await self.post(
      'uas/directLogout',
      data={
        'rememberMeOptIn': 'false',
        'JSESSIONID': self.get_csrf_token(),
        'lang': f'v=2&lang={self.config.linkedin_stored_data.li_lang}',
        'logout_reason': 'USER_INITIATED'
      },
      headers={
        'x-li-user-agent': f'LIAuthLibrary:0.0.3 {self.config.linkedin_app.appId}:{self.config.linkedin_app.clientVersion}'
                           f' {self.config.device.model}:{self.config.device.androidVersion}'
      }
    )



from typing import Optional

import aiohttp

from linkauto.api.mixins.stub import StubMixin


class ProfileMixin(StubMixin):
  async def profile_url_to_id(self, url: str) -> str:
    return url.lstrip('/').replace('in/', '')

  async def profile_get(self, profile_url: Optional[str] = None, id_: Optional[str] = None) -> aiohttp.ClientResponse:
    if not any([profile_url, id_]):
      raise ValueError("Neither an id, or an url has been provided.")

    if id_ is None:
      id_ = await self.profile_url_to_id(profile_url)

    return await self.get(
      f'/voyager/api/voyagerIdentityDashProfiles',
      page_instance='urn:li:page:p_flagship3_profile_view_base',
      accept_json=True,
      query_params={
        'memberIdentity': id_,
        'q': 'memberIdentity',
        'decorationId': 'com.linkedin.voyager.dash.deco.identity.profile.LocalizedProfileWithEntities-68',
      }
    )

  async def profile_get_recommendations(self, given: bool = False, profile_url: Optional[str] = None, id_: Optional[str] = None) -> aiohttp.ClientResponse:
    if not any([profile_url, id_]):
      raise ValueError("neither an id, or an url has been provided.")

    if id_ is None:
      id_ = await self.profile_url_to_id(profile_url)

    return await self.get(
      f'voyager/api/identity/profiles/{id_}/recommendations',
      page_instance='urn:li:page:p_flagship3_profile_view_base',
      accept_json=True,
      query_params={
        'q': 'given' if given else 'received',
      }
    )

  async def profile_get_skills(self, profile_url:  Optional[str] = None, id_: Optional[str] = None) -> aiohttp.ClientResponse:
    if not any([profile_url, id_]):
      raise ValueError("neither an id, or an url has been provided.")

    if id_ is None:
      id_ = await self.profile_url_to_id(profile_url)

    return await self.get(
      f'voyager/api/identity/profiles/{id_}/skills',
      page_instance='urn:li:page:p_flagship3_profile_view_base',
      accept_json=True,
    )

  async def profile_get_connections(self, in_common: bool = False, profile_url:  Optional[str] = None, id_: Optional[str] = None) -> aiohttp.ClientResponse:
    if not any([profile_url, id_]):
      raise ValueError("neither an id, or an url has been provided.")

    if id_ is None:
      id_ = await self.profile_url_to_id(profile_url)

    qp = {'q': 'inCommon' if in_common else 'connections'}
    if not in_common:
      qp['count'] = 10

    return await self.get(
      f'voyager/api/identity/profiles/{id_}/memberConnections',
      page_instance='urn:li:page:p_flagship3_profile_view_base',
      accept_json=True,
      query_params=qp
    )

  async def profile_get_following(self, profile_url:  Optional[str] = None, id_: Optional[str] = None) -> aiohttp.ClientResponse:
    if not any([profile_url, id_]):
      raise ValueError("neither an id, or an url has been provided.")

    if id_ is None:
      id_ = await self.profile_url_to_id(profile_url)

    return await self.get(
      f'voyager/api/identity/profiles/{id_}/following',
      page_instance='urn:li:page:p_flagship3_profile_view_base',
      accept_json=True,
    )

  async def profile_get_contact_info(self, profile_url: Optional[str] = None,
                                    id_: Optional[str] = None) -> aiohttp.ClientResponse:
    if not any([profile_url, id_]):
      raise ValueError("neither an id, or an url has been provided.")

    if id_ is None:
      id_ = await self.profile_url_to_id(profile_url)

    return await self.get(
      f'voyager/api/identity/profiles/{id_}/profileContactInfo',
      page_instance='urn:li:page:p_flagship3_profile_view_base',
      accept_json=True
    )



from linkauto.api.dataclasses_ import Device, LinkedinApp

GooglePixelUS = Device(
  osName='Android OS',
  osVersion='25',
  androidVersion='Android 7.1.1',
  region='en_US',
  cronetVersion='83.0.4103.83',
  model='unknown_Google Pixel XL',
  displayDensity=3.5,
  displayWidth=1440,
  displayHeight=2392,
  dpi='xhdpi',
  deviceType='android',
  timezoneOffset=-4,
  timezone='America/New_York',
  isAdTrackingLimited=False,
)

Linkedin_41588_146600 = LinkedinApp(
  clientVersion='4.1.558',
  clientMinorVersion=146600,
  appId='com.linkedin.android',
  storeId='us_googleplay',
  mpName='voyager-android',
  mpVersion='0.6.34.153'
)



import asyncio
from typing import Type

from linkauto.api.constants import Linkedin_41588_146600, GooglePixelUS
from linkauto.api.dataclasses_ import Device, LinkedinStoredData, LinkedinApp, Config
from linkauto.api.mixins.auth import AuthMixin
from linkauto.api.mixins.helper import HelperMixin
from linkauto.api.mixins.http import HttpMixin
from linkauto.api.mixins.profile import ProfileMixin
from linkauto.api.mixins.relationship import RelationshipMixin
from linkauto.api.mixins.track import TrackMixin


class ApiClient(HttpMixin, TrackMixin, AuthMixin, HelperMixin, ProfileMixin, RelationshipMixin):
  def __init__(self, username: str, password: str, linkedin_app: Type[LinkedinApp], device: Type[Device]):
    self.config = Config(device, linkedin_app, LinkedinStoredData(device.region))
    self.username = username
    self.password = password
    self.start_http_mixin()

  async def init(self):
    await self.start_track_mixin()

  async def stop(self):
    await self.stop_http_mixin()




import dataclasses
import uuid
from typing import Type


@dataclasses.dataclass
class Device:
  osName: str
  osVersion: str
  model: str
  displayDensity: float
  displayWidth: int
  displayHeight: int
  dpi: str
  deviceType: str
  timezoneOffset: int
  timezone: str
  isAdTrackingLimited: bool
  androidVersion: str
  region: str
  cronetVersion: str
  deviceId: str = None


@dataclasses.dataclass
class LinkedinApp:
  clientVersion: str
  clientMinorVersion: int
  storeId: str
  appId: str
  mpName: str
  mpVersion: str


@dataclasses.dataclass
class LinkedinStoredData:
  li_lang: str
  li_udid: uuid.UUID = dataclasses.field(default_factory=lambda: str(uuid.uuid4))
  li_fabric: str = None
  li_pop: str = None
  li_proto: str = None
  li_uuid: str = None


@dataclasses.dataclass
class Config:
  device: Type[Device]
  linkedin_app: Type[LinkedinApp]
  linkedin_stored_data: LinkedinStoredData




import setuptools
from distutils.core import setup

setup(
  name='linkauto',
  packages=setuptools.find_packages(),
  version='0.0.4',
  license='MIT',
  description='Python wrapper for the private LinkedIn API',
  author='Stan van Rooy',
  author_email='stan@rooy.dev',
  url='https://github.com/stanvanrooy/linkauto',
  download_url='https://github.com/stanvanrooy/linkauto/archive/0.0.1.tar.gz',
  keywords=['linkedin api', 'private linkedin api'],
  install_requires=[
    'orjson',
    'aiohttp'
  ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.8',
  ],
)



import aiohttp
import orjson
import asyncio
import aioscheduler
import random
import datetime

from linkauto.api.mixins.stub import StubMixin

class TrackMixin(StubMixin):
  def __init__(self) -> None:
    self._scheduler = aioscheduler.TimedScheduler()
    self._scheduler.start()
    self._schedule_timestamp_event()

  async def start_track_mixin(self):
    # 1. This request retrieves some information necessary for all track requests.
    await self.get(
      'li/track',
      headers={'user-agent': self.get_useragent()},
      default_headers=False
    )

    # 2. Retrieve useless configuration
    await self.get(
      'voyager/api/configuration',
      headers={
        'x-li-track': self.get_track_header(),
        'x-li-lang': self.config.linkedin_stored_data.li_lang,
        'accept': 'application/vnd.linkedin.mobile.deduped+json',
        'user-agent': self.get_useragent(),
        'x-li-page-instance': self.get_page_instance('p_flagship3_background'),
        'csrf-token': self.get_csrf_token(),
        'x-udid': str(self.config.linkedin_stored_data.li_udid),
        'x-restli-protocol-version': '2.0.0'
      }
    )
    
    self.add_response_interceptor(self.set_tracking_headers)
    self.add_response_interceptor(self._generate_network_event)


  async def _generate_network_event(self, response: aiohttp.ClientResponse):
    # TODO: implement
    pass

  async def generate_rum_event(self):
    # TODO: implement
    pass

  async def set_tracking_headers(self, response: aiohttp.ClientResponse):
    headers = response.headers
    self.config.linkedin_stored_data.li_pop = headers.get('x-li-pop', self.config.linkedin_stored_data.li_pop)
    self.config.linkedin_stored_data.li_uuid = headers.get('x-li-uuid', self.config.linkedin_stored_data.li_uuid)
    self.config.linkedin_stored_data.li_proto = headers.get('x-li-proto', self.config.linkedin_stored_data.li_proto)
    self.config.linkedin_stored_data.li_fabric = headers.get('x-li-fabric', self.config.linkedin_stored_data.li_fabric)

  def get_track_header(self) -> str:
    x_li_track = {
      "osName": self.config.device.osName,
      "osVersion": self.config.device.osVersion,
      "model": self.config.device.model,
      "displayDensity": self.config.device.displayDensity,
      "displayWidth": self.config.device.displayWidth,
      "displayHeight": self.config.device.displayHeight,
      "dpi": self.config.device.dpi,
      "deviceType": self.config.device.deviceType,
      "deviceId": self.config.device.deviceId,
      "timezoneOffset": self.config.device.timezoneOffset,
      "timezone": self.config.device.timezone,
      "isAdTrackingLimited": self.config.device.isAdTrackingLimited,
      "appId": self.config.linkedin_app.appId,
      "storeId": self.config.linkedin_app.storeId,
      "clientVersion": self.config.linkedin_app.clientVersion,
      "clientMinorVersion": self.config.linkedin_app.clientMinorVersion,
      "mpName": self.config.linkedin_app.mpName,
      "mpVersion": self.config.linkedin_app.mpVersion
    }
    return orjson.dumps(x_li_track).decode('utf8')

  async def generate_network_event(self, _: aiohttp.ClientResponse):
    # TODO: implement
    pass

  async def _generate_timestamp_event(self) -> None:
    for _ in range(10):
        url = "https://realtime.www.linkedin.com/realtime/realtimeFrontendTimestamp"
        headers = {
          'x-li-lang': self.config.linkedin_stored_data.li_lang
        }
        _ = self.get(url, headers=headers)
        await asyncio.sleep(.750)

  def _schedule_timestamp_event(self) -> None:
    self._scheduler.schedule(
      self._generate_timestamp_event(), 
      datetime.datetime.utcnow() + datetime.timedelta(seconds=random.randint(100, 150))
    )

